#!/usr/bin/env node
'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var fs = _interopDefault(require('fs'));
var path = require('path');
var chalk = require('chalk');

function __async(g){return new Promise(function(s,j){function c(a,x){try{var r=g[x?"throw":"next"](a);}catch(e){j(e);return}r.done?s(r.value):Promise.resolve(r.value).then(c,d);}function d(e){c(e,1);}c();})}

class Hookfinder {
  constructor (options = {}) {
    this.targets = options.targets;
    this.basefolder = options.basefolder;
  }

  start () {return __async(function*(){
    const result = yield Promise.all(Object.keys(this.targets).map(classPath => __async(function*(){
      const filePath = path.join(this.basefolder, classPath.replace(/\./g, '/') + '.smali');
      const source = yield this.readFile(filePath);
      const methods = (this.targets[classPath].methods || []).map(method => {
        let implementations = source
                    // match entire method
                    .match(new RegExp(method.filter.source + /(.|\n|\r)*?\.end method/.source, 'g'))
                    // use implementation filter if defined
                    .filter(impl => {
                      if (typeof method.implementationFilter === 'undefined') return true
                      return impl.match(method.implementationFilter) !== null
                    });
        let names = implementations.map(impl => this.getCapture(impl, method.filter));

        return {
          name: method.name,
          matches: names,
          implementations: implementations
        }
      });
      const fields = (this.targets[classPath].fields || []).map(field => {
        let fieldNames = source.match(field.filter)
                    .map(smali => this.getCapture(smali, field.filter));
        return {
          name: field.name,
          matches: fieldNames
        }
      });

      return {
        name: classPath,
        methods,
        fields
      }
    }.call(this))));
    result.forEach(r => {
      console.log('#', r.name);
      const {methods, fields} = r;
      if (methods.length) {
        console.log('## Methods');
        methods.forEach(m => {
          console.log(m.name, '=', m.matches.join(','));
          if (m.matches.length > 1) {
            console.log(chalk.yellow(m.implementations.join('\n')));
          }
        });
      }
      if (fields.length) {
        console.log('## Fields');
        fields.forEach(f => console.log(f.name, '=', f.matches.join(',')));
      }
      console.log();
    });
  }.call(this))}

  getCapture (source, regex) {
    return source.match(new RegExp(regex.source))[1]
  }

  readFile (filePath) {
    return new Promise((resolve, reject) => {
      fs.readFile(filePath, 'utf8', (err, data) => {
        if (err) return reject(err)
        resolve(data);
      });
    })
  }
}

var version = "0.0.3";

const cmd = require('commander');

cmd
  .version(version)
  .option('-s, --settings <settings-file>', 'Specify settings')
  .option('-v, --verbose', 'Show stacktraces');

cmd
  .command('find <base-folder>')
  .description('search in the specified directory')
  .action((basefolderOpt) => {
    let basefolder, targets;
    try {
      const settings = require(path.join(process.cwd(), cmd.settings || 'hookfinder.settings'));
      basefolder = path.join(basefolderOpt, settings.PACKAGE_BASE);
      targets = settings.targets;
    } catch (err) {
      console.error('Failed to load settings file');
      if (cmd.verbose) console.error(err);
      return
    }
    new Hookfinder({
      basefolder: basefolder,
      targets
    }).start()
      .catch(err => {
        console.error(err);
      });
  });

cmd.parse(process.argv);
